# -*- encoding: utf-8 -*-
#
# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import getpass
import os

from builtins import input
from configparser import ConfigParser


def do_genconfig(args):
    """Generates the config file according to user input"""

    print("========= DEFAULT ========")
    retry_interval = input("Type the polling interval in seconds for "
                           "daemon to manage the nodes "
                           "e.g., 300 for every 5 minutes: ")
    retry_interval = retry_interval if retry_interval else "300"
    rpc_thread_pool_size = input("Type the size of the RPC thread pool: ")
    rpc_thread_pool_size = \
        rpc_thread_pool_size if rpc_thread_pool_size else "20"

    print("========= Openstack ========= ")
    openstack_auth_url = input("Type the auth_url for the Ironic service: ")
    openstack_username = input("Type your OpenStack username: ")
    openstack_tenant = input("Type your OpenStack user's tenant name: ")
    openstack_password = getpass.getpass(
        "Type your Openstack user's password: "
    )
    openstack_insecure = input("Would you like to allow insecure connections "
                               "to OpenStack? [y/N]: ") or "N"
    openstack_insecure = 'True' if openstack_insecure.lower() == 'y' \
        else 'False'
    default_deploy_kernel = input("Type in the default deploy keynel image "
                                  "ID on Glance: ")
    default_deploy_ramdisk = input("Type in the default deploy ramdisk "
                                   "image ID on Glance: ")

    # TODO(thiagop): get drivers enabled with ironicclient
    enabled_drivers = ['agent_pxe_oneview', 'iscsi_pxe_oneview']
    ironic_default_driver = input(("Which driver would you like to use? "
                                   "[%s]: ") % ','.join(enabled_drivers))

    print("========= OneView ========= ")
    oneview_manager_url = input("Type in the OneView uri: ")
    oneview_username = input("Type your OneView username: ")
    oneview_password = getpass.getpass("Type your OneView user's password: ")
    oneview_insecure = input("Would you like to allow insecure connections "
                             "to OneView? [y/N]: ") or "N"
    oneview_insecure = 'True' if oneview_insecure.lower() == 'y' else 'False'

    config = ConfigParser()
    config.set("DEFAULT", "retry_interval", retry_interval)
    config.set("DEFAULT", "rpc_thread_pool_size", rpc_thread_pool_size)

    config.add_section("openstack")
    config.set("openstack", "auth_url", openstack_auth_url)
    config.set("openstack", "admin_user", openstack_username)
    config.set("openstack", "admin_tenant_name", openstack_tenant)
    config.set("openstack", "admin_password", openstack_password)
    config.set("openstack", "insecure", openstack_insecure)
    config.set("openstack", "default_deploy_kernel_id", default_deploy_kernel)
    config.set(
        "openstack", "default_deploy_ramdisk_id", default_deploy_ramdisk
    )
    config.set("openstack", "default_driver", ironic_default_driver)

    config.add_section("oneview")
    config.set("oneview", "manager_url", oneview_manager_url)
    config.set("oneview", "username", oneview_username)
    config.set("oneview", "password", oneview_password)
    config.set("oneview", "allow_insecure_connections", oneview_insecure)

    filename = input("Type the path of the new configuration file [%s]: "
                     % args.config_file) or args.config_file
    full_filename = os.path.realpath(os.path.expanduser(filename))
    directory = os.path.dirname(full_filename)

    if not os.path.exists(directory):
        os.makedirs(directory)

    with open(full_filename, 'w') as configfile:
        config.write(configfile)
        print("======\nFile created successfully on '%s'!\n======" % filename)



# -*- encoding: utf-8 -*-
#
# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import configparser as ConfigParser

import os

import ironic_oneviewd.service_logging as logging


LOG = logging.getLogger(__name__)


class ConfClient:
    def __init__(self, configname, defaults={}):
        self._CONF = ConfigParser.ConfigParser(defaults)
        path = os.path.realpath(os.path.expanduser(configname))
        self._CONF.readfp(open(path))

    def __getattr__(self, section):

        class Section(object):
            def __init__(self, cfg_section):
                self.__dict__ = dict(cfg_section)

            def __getattribute__(self, *args, **kwargs):
                try:
                    return object.__getattribute__(self, *args, **kwargs)
                except AttributeError:
                    raise AttributeError("Missing required attribute '%s' on "
                                         "section '%s' " % (args[0], section))

        ret = Section(self._CONF.items(section))
        return ret



# -*- encoding: utf-8 -*-
#
# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic_oneviewd import utils
from oneview_client import exceptions as oneview_exceptions


class Facade(object):

    def __init__(self, config):
        self.ironicclient = utils.get_ironic_client(config)
        self.oneview_client = utils.get_oneview_client(
            config.oneview.manager_url,
            config.oneview.username,
            config.oneview.password,
            config.oneview.allow_insecure_connections,
            config.oneview.tls_cacert_file,
        )

    # =========================================================================
    # Ironic actions
    # =========================================================================

    def get_ironic_node_list(self):
        return self.ironicclient.node.list(detail=True)

    def get_ironic_node(self, node_uuid):
        return self.ironicclient.node.get(node_uuid)

    def delete_ironic_node(self, node_uuid):
        return self.ironicclient.node.delete(node_uuid)

    def node_set_maintenance(self, node_uuid, maintenance_mode, maint_reason):
        return self.ironicclient.node.set_maintenance(
            node_uuid,
            maintenance_mode,
            maint_reason=maint_reason
        )

    def create_ironic_node(self, **attrs):
        return self.ironicclient.node.create(**attrs)

    def set_node_provision_state(self, node, state):
        return self.ironicclient.node.set_provision_state(node.uuid, state)

    def get_port_list_by_node_uuid(self, node_uuid):
        return self.ironicclient.port.get(node_id=node_uuid)

    def get_port(self, port_id):
        return self.ironicclient.port.get(port_id)

    def get_port_list_by_mac(self, port_mac_address):
        return self.ironicclient.port.list(address=port_mac_address)

    def create_node_port(self, node_uuid, port_mac_address):
        return self.ironicclient.port.create(node_uuid=node_uuid,
                                             address=port_mac_address)

    # =========================================================================
    # OneView actions
    # =========================================================================

    def get_server_hardware(self, node_info):
        return self.oneview_client.get_server_hardware(node_info)

    def get_server_hardware_state(self, uuid):
        return self.oneview_client.get_server_hardware_state(
            uuid
        )

    def get_server_profile_assigned_to_sh(self, node_info):
        return self.oneview_client.get_server_profile_from_hardware(
            node_info
        )

    def generate_and_assign_sp_from_spt(
        self, server_profile_name, node_info):
        return self.oneview_client.apply_server_profile(
            server_profile_name, node_info
        )

    def unassign_server_profile(self, node_info):
        return self.oneview_client.delete_server_profile_from_server_hardware(
            node_info
        )

    def get_server_hardware_mac(self, uuid):
        sh = self.oneview_client.get_server_hardware_by_uuid(uuid)
        try:
            # MAC from ServerHardware
            mac = sh.get_mac(nic_index=0)
        except oneview_exceptions.OneViewException:
            # MAC from iLO
            mac = self.oneview_client.get_sh_mac_from_ilo(sh.uuid)
        return mac



# -*- encoding: utf-8 -*-
#
# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironicclient import client as ironic_client

from oslo_utils import importutils

from ironic_oneviewd.openstack.common._i18n import _
from ironic_oneviewd import exceptions
from ironic_oneviewd import service_logging as logging

client = importutils.try_import('oneview_client.client')
oneview_states = importutils.try_import('oneview_client.states')
oneview_exceptions = importutils.try_import('oneview_client.exceptions')

REQUIRED_ON_PROPERTIES = {
    'server_hardware_type_uri': _("Server Hardware Type URI Required."),
    'server_profile_template_uri': _("Server Profile Template URI required"),
}

REQUIRED_ON_EXTRAS = {
    'server_hardware_uri': _("Server Hardware URI. Required."),
}

IRONIC_API_VERSION = '1.11'

LOG = logging.getLogger(__name__)


def get_ironic_client(conf):
    kwargs = {
        'os_username': conf.openstack.admin_user,
        'os_password': conf.openstack.admin_password,
        'os_auth_url': conf.openstack.auth_url,
        'os_tenant_name': conf.openstack.admin_tenant_name,
        'os_ironic_api_version': IRONIC_API_VERSION,
    }
    if conf.openstack.insecure.lower() == 'true':
        kwargs['insecure'] = True
    if conf.openstack.ca_file:
        kwargs['ca_file'] = conf.openstack.ca_file

    LOG.debug("Using OpenStack credentials specified in the configuration "
              "file to get Ironic Client")

    return ironic_client.get_client(1, **kwargs)


def get_oneview_client(manager_url, username, password,
                       allow_insecure_connections=False, tls_cacert_file='',
                       max_polling_attempts=20):
    """Generates an instance of the OneView client.

    Generates an instance of the OneView client using the imported
    oneview_client library.

    :returns: an instance of the OneView client
    """

    oneview_client = client.Client(
        manager_url=manager_url,
        username=username,
        password=password,
        allow_insecure_connections=allow_insecure_connections,
        tls_cacert_file=tls_cacert_file,
        max_polling_attempts=max_polling_attempts
    )
    return oneview_client


def verify_node_properties(node):
    properties = node.properties.get('capabilities', '')
    for key in REQUIRED_ON_PROPERTIES:
        if key not in properties:
            raise exceptions.MissingParameterValue(
                _("Missing the following OneView data in node's "
                  "properties/capabilities: %s.") % key
            )

    return properties


def verify_node_extra(node):
    extra = node.extra or {}
    for key in REQUIRED_ON_EXTRAS:
        if not extra.get(key):
            raise exceptions.MissingParameterValue(
                _("Missing the following OneView data in node's extra: %s.")
                % key
            )

    return extra


def capabilities_to_dict(capabilities):
    capabilities_dict = {}
    if capabilities:
        try:
            for capability in capabilities.split(','):
                key, value = capability.split(':')
                capabilities_dict[key] = value
        except ValueError:
            raise exceptions.InvalidParameterValue(
                _("Malformed capabilities value: %s") % capability
            )

    return capabilities_dict


def dynamic_allocation_enabled(node):
    return_value = False
    flag = node.driver_info.get('dynamic_allocation')
    if flag is not None:
        if flag in ('true', 'True', True):
            return_value = True
        else:
            error_msg = ("Invalid dynamic_allocation parameter value in "
                         "node's %(node_uuid)s driver_info. Valid values "
                         "are booleans true or false." %
                         {"node_uuid": node.uuid})
            LOG.error(error_msg)
            raise exceptions.InvalidParameterValue(_(error_msg))
    return return_value



# -*- encoding: utf-8 -*-
#
# Copyright 2015 Hewlett-Packard Development Company, L.P.
# Copyright 2015 Universidade Federal de Campina Grande
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six
import traceback

from concurrent import futures

from ironic_oneviewd import exceptions
from ironic_oneviewd import facade
from ironic_oneviewd.openstack.common._i18n import _
from ironic_oneviewd import service_logging as logging
from ironic_oneviewd import utils

LOG = logging.getLogger(__name__)


ENROLL_PROVISION_STATE = 'enroll'
MANAGEABLE_PROVISION_STATE = 'manageable'
ONEVIEW_PROFILE_APPLIED = 'ProfileApplied'

SUPPORTED_DRIVERS = ["agent_pxe_oneview",
                     "iscsi_pxe_oneview",
                     "fake_oneview"]


class NodeManager(object):

    def __init__(self, conf_client):

        self.facade = facade.Facade(conf_client)
        self.max_workers = int(
            conf_client.DEFAULT.rpc_thread_pool_size
        )
        self.executor = futures.ThreadPoolExecutor(
            max_workers=self.max_workers
        )

    def pull_ironic_nodes(self):
        ironic_nodes = self.facade.get_ironic_node_list()

        LOG.info(
            "%(nodes)s Ironic nodes has been taken." %
            {"nodes": len(ironic_nodes)}
        )

        nodes = [node for node in ironic_nodes
                 if node.driver in SUPPORTED_DRIVERS
                 if node.maintenance is False]

        self.executor.map(self.manage_node_provision_state, nodes)

    def manage_node_provision_state(self, node):
        if node.provision_state == ENROLL_PROVISION_STATE:
            self.take_enroll_state_actions(node)
        elif node.provision_state == MANAGEABLE_PROVISION_STATE:
            self.take_manageable_state_actions(node)

    def take_enroll_state_actions(self, node):
        LOG.info(
            "Taking enroll state actions for node %(node)s." %
            {'node': node.uuid}
        )

        if not utils.dynamic_allocation_enabled(node):
            try:
                self.apply_server_profile(
                    node
                )
            except \
                exceptions.NodeAlreadyHasServerProfileAssignedException as ex:
                LOG.warning(six.text_type(ex))
            except exceptions.ServerProfileApplicationException as ex:
                LOG.warning(six.text_type(ex))
                return

            try:
                self.apply_node_port_configuration(
                    node
                )
            except exceptions.NodeAlreadyHasPortForThisMacAddress as ex:
                LOG.warning(six.text_type(ex))
            except exceptions.NoBootableConnectionFoundException as ex:
                LOG.warning(six.text_type(ex))
                return
        else:
            try:
                self.apply_node_port_conf_for_dynamic_allocation(
                    node
                )
            except exceptions.NodeAlreadyHasPortForThisMacAddress as ex:
                LOG.warning(six.text_type(ex))
            except exceptions.NoBootableConnectionFoundException as ex:
                LOG.warning(six.text_type(ex))
                return

        try:
            self.facade.set_node_provision_state(node, 'manage')
        except Exception:
            LOG.error(traceback.format_exc())
            return

    def take_manageable_state_actions(self, node):
        LOG.info(
            "Taking manageable state actions for node %(node)s." %
            {'node': node.uuid}
        )

        try:
            self.facade.set_node_provision_state(node, 'provide')
        except Exception:
            LOG.error(traceback.format_exc())

    def server_hardware_has_server_profile_fully_applied(self, node):
        server_hardware_uuid = self.server_hardware_uuid_from_node(node)
        server_hardware_state = self.facade.get_server_hardware_state(
            server_hardware_uuid
        )
        return server_hardware_state == ONEVIEW_PROFILE_APPLIED

    def server_hardware_has_server_profile_applied(self, node):
        server_hardware_uri = self.server_hardware_uri_from_node(
            node
        )
        profile_applied = \
            self.facade.is_server_profile_applied_on_server_hardware(
                server_hardware_uri
            )
        return profile_applied

    def apply_server_profile(self, node):
        server_profile_uri = self.server_profile_uri_from_node(
            node
        )

        node_info = self.get_node_info_from_node(node)

        server_profile_name = "Ironic [%s]" % (node.uuid)

        if server_profile_uri is None:
            try:
                server_profile_uri = \
                    self.facade.generate_and_assign_sp_from_spt(
                        server_profile_name,
                        node_info
                    )
            except Exception:
                raise exceptions.ServerProfileApplicationException(node)
        else:
            raise exceptions.NodeAlreadyHasServerProfileAssignedException(
                node,
                server_profile_uri
            )

    def apply_node_port_configuration(self, node):
        if not self.server_hardware_has_server_profile_fully_applied(node):
            LOG.error(
                "The Server Profile application is not "
                "finished yet for node %(node)s." %
                {'node': node.uuid}
            )
            return

        node_info = self.get_node_info_from_node(node)

        assigned_server_profile_uri = self.server_profile_uri_from_node(
            node
        )
        server_profile = self.facade.get_server_profile_assigned_to_sh(
            node_info
        )
        primary_boot_connection = None

        if server_profile.connections:
            for connection in server_profile.connections:
                boot = connection.get('boot')
                if (boot is not None and
                   boot.get('priority').lower() == 'primary'):
                    primary_boot_connection = connection

            if primary_boot_connection is None:
                raise exceptions.NoBootableConnectionFoundException(
                    assigned_server_profile_uri
                )
            mac = primary_boot_connection.get('mac')
        else:
            server_hardware_uuid = self.server_hardware_uuid_from_node(node)
            mac = self.facade.get_server_hardware_mac(server_hardware_uuid)

        port_list_by_mac = self.facade.get_port_list_by_mac(mac)

        if not port_list_by_mac:
            return self.facade.create_node_port(node.uuid, mac)
        else:
            port_obj = self.facade.get_port(port_list_by_mac[0].node_uuid)
            if port_obj.node_uuid != node.uuid:
                return self.facade.create_node_port(
                    node.uuid, mac
                )
            else:
                raise exceptions.NodeAlreadyHasPortForThisMacAddress(
                    mac
                )

    def apply_node_port_conf_for_dynamic_allocation(self, node):
        server_hardware_uuid = self.server_hardware_uuid_from_node(node)
        mac = self.facade.get_server_hardware_mac(server_hardware_uuid)

        port_list_by_mac = self.facade.get_port_list_by_mac(mac)

        if not port_list_by_mac:
            return self.facade.create_node_port(node.uuid, mac)
        else:
            port_obj = self.facade.get_port(port_list_by_mac[0].node_uuid)
            if port_obj.node_uuid != node.uuid:
                return self.facade.create_node_port(
                    node.uuid, mac
                )
            else:
                raise exceptions.NodeAlreadyHasPortForThisMacAddress(
                    mac
                )

    def server_hardware_uri_from_node(self, node):
        return node.driver_info.get(
            'server_hardware_uri'
        )

    def server_profile_template_uri_from_node(self, node):
        node_capabilities = self.capabilities_to_dict(
            node.properties.get('capabilities')
        )
        node_server_profile_template_uri = node_capabilities.get(
            'server_profile_template_uri'
        )
        return node_server_profile_template_uri

    def server_profile_uri_from_node(self, node):
        node_info = self.get_node_info_from_node(node)
        server_profile_uri = None
        try:
            server_profile = self.facade.get_server_profile_assigned_to_sh(
                node_info
            )

            if node.uuid in server_profile.name:
                return server_profile.uri
            else:
                return server_profile_uri

        except Exception:
            return server_profile_uri

    def get_node_info_from_node(self, node):
        capabilities_dict = self.capabilities_to_dict(
            node.properties.get('capabilities', '')
        )
        driver_info = node.driver_info
        oneview_info = {
            'server_hardware_uri':
                driver_info.get('server_hardware_uri'),
            'server_hardware_type_uri':
                capabilities_dict.get('server_hardware_type_uri'),
            'enclosure_group_uri':
                capabilities_dict.get('enclosure_group_uri'),
            'server_profile_template_uri':
                capabilities_dict.get('server_profile_template_uri') or
                driver_info.get('server_profile_template_uri')
        }
        return oneview_info

    def uuid_from_uri(self, uri):
        return uri.split("/")[-1]

    def server_hardware_uuid_from_node(self, node):
        uri = self.server_hardware_uri_from_node(node)
        return self.uuid_from_uri(uri)

    def capabilities_to_dict(self, capabilities):
        """Parse the capabilities string into a dictionary

        :param capabilities: the node capabilities as a formatted string
        :raises: InvalidParameterValue if capabilities is not an string or has
                 a malformed value
        """
        capabilities_dict = {}
        if capabilities:
            if not isinstance(capabilities, six.string_types):
                raise exceptions.InvalidParameterValue(
                    _("Value of 'capabilities' must be string. Got %s")
                    % type(capabilities))
            try:
                for capability in capabilities.split(','):
                    key, value = capability.split(':')
                    capabilities_dict[key] = value
            except ValueError:
                raise exceptions.InvalidParameterValue(
                    _("Malformed capabilities value: %s") % capability
                )
        return capabilities_dict
